package provide MTClient 0.1

package require Thread
package require uuid

##
# MTClient
# 	клиент для MTWorker(s)
#
#   конструктор без аргументов
#            MTClient create mtc
#   set mtc [ MTClient new ]
#
# узнать собственный id
#   mtc id 
#
# ожидать готовность воркеров, вернёт кол-во воркеров
#   mtc workers ?msc?
#
# послать команду воркерам, получить id запроса (promise)
# msc - таймер ожидания ответа, милисек
# 	mtc async msc command ?args..? 
#
# послать несколько команд воркерам
#   mtc asyncmany msc commandlist 
#
# проверить готовность запроса, без ожидания
# 	mtc complete $promise
#
# получить результат, ожидать не более msс миллисекунд
#	mtc await ?-timeout msc? $promise
#
# "забыть" про promise, просто вычеркнуть из списков
#   mtc forget promise ...
#   mtc forget promise_list
#
# ожидать завершения одного из promise, вернёт первый который завершён
#   mtc any ?-nothrow? ?-timeout msc? promise ...
#   mtc any ?-nothrow? ?-timeout msc? promise_list
#
# ожидать завершения всех, вернёт cписок завершённых
#   mtc all ?-nothrow? ?-timeout msc? promise ...
#   mtc all ?-nothrow? ?-timeout msc? promise_list
# 
##
# для использования совместно с MetaTrader, promise - это просто числа(идентификаторы), не объёкты
#
# не требует event-loop, то есть вызовов update,update idletasks
#
# таймауты msc могут задаваться как относительные в милисек, то есть сколько милисек. можно ждать
# так и абсолютые [ clock millisecods ]+милисек, то есть до какого времени можно ждать
#
# при истечении таймаута команды вызывается "исключение" (код ошибки TCL_ERROR), а в TclResult() пусто
# при истечении таймаута promise тоже исключение (код TCL_ERROR), а в TclResult - только целое число=promise
# при всех прочих ошибках в TclResult описания ошибок
#
# методов с бесконечным ожиданием нет, таймаут равный 0 - для async значит не отслеживать экспирацию исполнения
# для остальных - однократное выполнение.   
##
oo::class create MTClient {
variable GUID	; # уникальный GUID
variable TSVAR	; # имя переменной tsv
variable PROMISE ; # промисы id=>dict
variable COUNTER ; # счётчик запросов 
variable LOCK 	;	
variable WAKEUP	;


variable WORKER_WAKEUP ; # 

variable startTime;	# время старта в секундах 

##
# конструктор без аргументов
# создаёт в tsv структуру $TSVAR:
# (далее - TSVAR - уникальное имя клиента)
#   incoming - входящая очередь (сюда вореры кладуд ответы)
#   wakeup   - thread::cond о приходе сообщения
#	lock     - thread::mutex блокировка
##
constructor {} {
	set startTime [ clock seconds ]
	set GUID 	[ uuid::uuid generate ]
	if { [ info exists ::MetaTrader(ProgramName) ] } {
		set TSVAR	$::MetaTrader(ProgramName)/$GUID
	} else {
		set TSVAR	"MTClient/$GUID"
	}
	array set PROMISE {}
	set LOCK 	[ thread::mutex create ]
	set WAKEUP 	[ thread::cond create ]
	set SHUTDOWN 0
	tsv::lock $TSVAR {
		tsv::set $TSVAR incoming {}
		tsv::set $TSVAR lock $LOCK
		tsv::set $TSVAR wakeup $WAKEUP
	}
	set WORKER_WAKEUP ""
	if { [ tsv::exists MTWorkers wakeup ] } {
		catch { set WORKER_WAKEUP [ tsv::get MTWorkers wakeup ] }
	}
	my print "hello !"
}
##
# деструктор - удаляет структуру в tsv
##
destructor {
	catch { thread::cond destroy $WAKEUP }
	catch { thread::mutex destroy $LOCK }
	catch { tsv::unset $TSVAR }
}
##
# id - собственный идентификатор
##
method id {} {
	set TSVAR
}

##
# parseopt timestamp arglist
#
# разбор типичной строки аргументов в словарик
#  timestamp - метка времени от которой отсчитывать отн.таймауты
#  arglist - параметры
# на выходе словарик:
#  timeout  absmcs - таймаут в абсолютном времени (для опции -timeout mcs)
#  nothrow   0|1   - разрешить/запретить бросать ошибки
#  userargs args  - копия переданного args
#  args  	{command ...} - оставшиеся аргументы, уже без опций -timeout и проч.
#  command  команда - первый элемент args 
##
# максимально просто
##
method parseopt { timestamp arglist } {
	set options [ list request {} userargs $arglist ]
	if { [ lindex $arglist 0 ] == "-nothrow" } {
		lappend options "-nothrow" 1
		set arglist [ lrange $arglist 1 end ]
	} else {
		lappend options "-nothrow" 0
	}	
	set command [ lindex $arglist 0 ]
	if { $command == {-timeout} } {
		set timeout [ lindex $arglist 1 ]
		if { $timeout == {} || ! [ string is int $timeout ] } {
			set timeout 0
		}
		if { $timeout > 0 &&  $timeout < $startTime } {
			set timeout [ expr { $timestamp + $timeout } ]
		}
		lappend options timeout $timeout args [ lrange $arglist 2 end ] command [ lindex $arglist 2 ]
	} else {
		lappend options timeout 0 args $arglist command $command
	}
	set options
}
##
# ready ?-timeout mcs?
# проверить наличие и состояние MTWorkers
# вернёт кол-во воркеров или 0 если нет ничего
# side-effect: может выставить/убрать/поменять WORKER_WAKEUP
##
method ready { args } {
	set ready 0
	if { [ catch {
		set ready [ tsv::get MTWorkers total ]
		if { $WORKER_WAKEUP == "" } {
			set WORKER_WAKEUP [ tsv::get MTWorkers wakeup ]
		}
	} err ] } {
		set WORKER_WAKEUP ""
		if { $args == 0 } {
			set timeout 0
		} else {
			set options [ my parseopt [ clock milliseconds ] $args ]
			set timeout [ dict get $options timeout ]
		}
		set msc 0
		while { $timeout != 0 && [ set msc [ expr { $timeout - [ clock milliseconds ] } ] ] > 0 } {
			after 5
			catch {
				set ready [ tsv::get MTWorkers total ]
				if { $WORKER_WAKEUP == "" } {
					set WORKER_WAKEUP [ tsv::get MTWorkers wakeup ]
				}
			}
			if { $ready>0 && $WORKER_WAKEUP != "" } {
				break
			}
		}
	}
	set ready
}
##
# async mcs command args 
# начать асинхронное (паралельное) исполнение
# буквально - передать команду воркерам, вернуть id
# timeout - задаёт макс.время на исполнение команды. По его истечению будет генериться ошибка
#
##
method async { timeout args } {
	# разбираем опции
	if { $timeout != 0 } {
		if { $timeout < $startTime } {
			set timeout [ expr { [ clock milliseconds ] + $timeout } ]
		}
	}
	if { [ llength $args ] == 1 } {
		set args [ lindex $args 0 ]
	}
	#set options [ my parseopt [ clock milliseconds ] $args ]
	#set timeout [ dict get $options timeout ]
	# получаем инкрементный номер 
	set id [ tsv::incr MTWorkers counter ]
	if { $id < 10 } {
		return -code error "no have workers"
	}
	# формируем запрос	
	set PROMISE($id) [ list "request" $args timeout $timeout ]
	# помещаем в очередь : {timeout 1 sender msgid команда на исполнение...}
	try {
		tsv::lpush MTWorkers incoming [ list $timeout 1 $TSVAR $id $args ]
	} on error err {
		return -code error "no have workers"
	}
	# оповещаем 
	try {
		thread::cond notify $WORKER_WAKEUP
	} on error code {
		# в теории, сервис мог успеть рестартовать
		set WORKER_WAKEUP ""
		catch { set WORKER_WAKEUP [ tsv::get MTWorkers wakeup ] }
		if { $WORKER_WAKEUP == "" } {
			unset PROMISE($id)
			return -code error "Workers not alived"
		}
		catch { thread::cond notify $WORKER_WAKEUP }
	}
	set id
}
##
# asyncmany msc commandlist
##
method asyncmany { msc commandlist } {
	# таймаут
	if { $timeout != 0 } {
		if { $timeout < $startTime } {
			set timeout [ expr { [ clock milliseconds ] + $timeout } ]
		}
	}
	# список запросов
	set count [ llength $commandlist ]
	if { $count == 0 } {
		return 0;
	}
	# получаем идентификаторы
	set last  [ tsv::incr MTWorkers counter $count ]
	set first [ expr { $last - $count } ]
	# готовим список запросов
	set requests {}
	set msgid $first
	foreach command $commandlist {
		lappend requests [ list $timeout 1 $TSVAR $msgid $command ]
		set PROMISE($msgid) [ list "request" $command timeout $timeout ]
		incr msgid
	}
	# отправляем
	tsv::lappend MTWorkers incoming {*}$requests
	# оповещаем 
	try {
		thread::cond notify $WORKER_WAKEUP
	} on error code {
		# в теории, сервис мог успеть рестартовать
		set WORKER_WAKEUP ""
		catch { set WORKER_WAKEUP [ tsv::get MTWorkers wakeup ] }
		if { $WORKER_WAKEUP == "" } {
			unset PROMISE($id)
			return -code error "Workers not alived"
		}
		catch { thread::cond notify $WORKER_WAKEUP }
	}
	# возвращаем кол-во
	set count
}
##
# invoke ?-timeout timeout? command args...
# "синхронный" вызов, аналог:  await -timeout $timeout [ async $timeout $command {*}args ]
# !! если таймаут не задан - будет ждать 30 секунд !!
##
method invoke { args } {
	tailcall my call {*}$args
}
method call { args } {
	set now [ clock milliseconds ]
	set options [ my parseopt $now $args ]
	set timeout [ dict get $options timeout ]
	set cmdline [ dict get $options args ]
	if { [ llength $cmdline ] == 1 } {
		set cmdline [ lindex $cmdline 0 ]
	}
	if { $timeout == 0 } {
		set timeout [ expr { $now + 30000 } ]
	}
	set promise [ my async $timeout $cmdline ]
	my await -timeout $timeout $promise  
}

##
# complete promise
# проверить на готовность результата (что await будет завершён без задержки)
# возможное исключение - promise отсутсвует в системе
# 0 - запрос ещё не выполнен
# 1 - выполнен успешно
# 2 - выполнен с ошибками
# ....
## 
method complete { id } {
	if { ! [ info exists PROMISE($id) ] } {
		return -code error "$id not found"
	}
	if { [ dict exists $PROMISE($id) error ] } {
		return 2
	}
	if { [ dict exists $PROMISE($id) result ] } {
		return 1
	}
	if { [ my loop ] > 0 } {
		if { ! [ info exists PROMISE($id) ] } {
			return -code error "$id not found"
		}
		if { [ dict exists $PROMISE($id) error ] } {
			return 2
		}
		if { [ dict exists $PROMISE($id) result ] } {
			return 1
		}
	}
	return 0
}
##
# await ?-timeout msc? promise
#
# ожидать завершения promise вернуть его результат
# при истечении таймаута генериться пустая ошибка
## 
method await { args } {
	set options [ my parseopt [ clock milliseconds ] $args ]
	set timeout [ dict get $options timeout ]
	set id [ lindex [ dict get $options command ] 0 ]
	if { [ llength $id ] != 1 || $id <=0 } {
		if { $id == {} } {
			return -code error "error"
		}
		return -code error $id
	}
	my loop
	while 1 {
		# обработать входящие
#		my loop $timeout
		# проверить состояние promise
		if { ! [ info exists PROMISE($id) ] } {
			return -code error $id
		}
		if { [ dict exists $PROMISE($id) error ] } {
			# исполнение завершилось с ошибкой
			# такую-же и сгенерить 
			set err [ dict get $PROMISE($id) error ]
			if { $err == {} } {
				# только пустую нельзя, чтобы не путать с таймаутом
				set err "error"
			}
			unset PROMISE($id)
			return -code error $err
		}
		if { [ dict exists $PROMISE($id) result ] } {
			# успешное завершение, отдать результат
			set result [ dict get $PROMISE($id) result ]
			unset PROMISE($id)
			return $result
		}
		# проверит таймаут сообщения
		set msgtimeout [ dict get $PROMISE($id) timeout ]
		if { $msgtimeout != 0 } {
			set now [ clock milliseconds ]
			if { $now>$msgtimeout } {
				# таймаут сообщения
				# ошибка с единственной цифрой
				unset PROMISE($id)
				return -code error $id
			}
		}
		if { $timeout > 0 } {
			if { $timeout < [ clock milliseconds ] } {
				return -code error {}
			}
		}
		if { [ catch {
			my loop $timeout
		} err ] } {
			#puts "loop error"
			return {}
		}
	}
}
##
# any ?-nothrow? ?-timeout msc? promise ...
# any ?-nothrow? ?-timeout msc? promise_list
#
# ожидать завершения любой задачи, вернуть её id (promise)
# гарантирует что последующий await будет без ожиданий
# или самому можно разобрать результат
# 
# -nothrow : не бросать собственные исключения
# -timeout msc : таймаут ожидания
# 
# если promise осутсвует, то в тексте ошибки будет только его номер
#      при -nothrow - будет возвращен 0 (сами разбирайтесь кого нет)
# если сработал таймаут, то текст ошибки будет пустым
#      при -nothrow - будет возвращено пустое значение {}
##
method any { args } {
	set options [ my parseopt [ clock milliseconds ] $args ]
	set timeout [ dict get $options timeout ]
	set nothrow [ dict get $options nothrow ]
	set ids [ dict get $options args ]
	if { [ llength $ids ] == 1 } {
		set ids [ lindex $ids 0 ]
	}
	set nexts $ids
	my loop
	while 1 {
		set waits $nexts
		set nexts {}
		foreach id $waits {
			if { $id <= 0 } {
				# некорректные id игнорируются
				continue
			}
			if { ! [ info exists PROMISE($id) ] } {
				if { $nothrow } {
					return 0
				}
				return -code error $id
			}
			if { [ dict exists $PROMISE($id) error ] } {
				#set err [ dict get $PROMISE($id) error ]
				#unset PROMISE($id)
				return -$id
			}
			if { [ dict exists $PROMISE($id) result ] } {
				set result [ dict get $PROMISE($id) result ]
				#unset PROMISE($id)
				return $id
			}
			set msgtimeout [ dict get $PROMISE($id) timeout ]
			if { $msgtimeout>0 && $msgtimeout < $now } {
				# таймаут сообщения, всё равно что его нет
				unset PROMISE($id)
				#dict set PROMISE($id) error $id
				if { $nothrow } {
					return 0
				}
				return -code error $id
			}
			lappend nexts $id
		}
		if { [ llength $nexts ] == 0 } {
			if { $nothrow } {
				return {}
			}
			return -code error 0
		}
		if { $timeout == 0 } {
			return {}
		}
		if { $timeout < [ clock milliseconds ] } {
			return {}
		}
		my loop $timeout
	}
}
##
# all ?-nothrow? ?-timeout msc? promise ...
# all ?-nothrow? ?-timeout msc? promise_list
#
# ожидать завершения всех указанных задач
# -nothrow : не выбрасывать собственных исключений
# -timeout msc : таймаут на ожидание
#
method all { args } {
	set options [ my parseopt [ clock milliseconds ] $args ]
	set timeout [ dict get $options timeout ]
	set nothrow [ dict get $options nothrow ]
	set ids [ dict get $options args ]
	if { [ llength $ids ] == 1 } {
		set ids [ lindex $ids 0 ]
	}
	set ready {}
	set nexts $ids
	my loop 
	while 1 {
		set waits $nexts
		set nexts {}
		foreach id $waits {
			if { $id == 0 } {
				# некорректые id игнорируются
				continue
			}
			if { $id < 0 } {
				lappend ready $id
				continue
			}
			if { ! [ info exists PROMISE($id) ] } {
				if { $nothrow } {
					return 0
				}
				return -code error $id
			}
			if { [ dict exists $PROMISE($id) error ] } {
				lappend ready -$id
			} elseif { [ dict exists $PROMISE($id) result ] } {
				lappend ready $id
			} else {
				set msgtimeout [ dict get $PROMISE($id) timeout ] 
				if { $msgtimeout > 0 && $msgtimeout < [ clock milliseconds ] } {
					# таймаут сообщения - всё равно что его нет
					unset PROMISE($id)
					if { $nothrow } {
						return 0
					}
					return -code error $id		
				}
				lappend nexts $id
			}
		}
		if { [ llength $nexts ] == 0 || [ llength $ready ] == [ llength $ids ] } {
			return $ready
		}
		if { $timeout == 0 } {
			return {}
		}
		if { $timeout < [ clock milliseconds ] } {
			if { $nothrow } {
				return {}
			}
			return -code error {}
		}
		set waits $nexts
		my loop $timeout
	}
}
method forget { ids } {
	set count 0
	foreach id $ids {
		if { [ info exists PROMISE($id) ] } {
			unset PROMISE($id)
			incr count
		}
		
	}
	return $count
}
method getmessage { { timeout 0 } } {
	set msg [ tsv::lpop $TSVAR incoming ]
	if { $msg != {} && $timeout > 0 } {
		set msc [ expr $timeout - [ clock milliseconds ] ] 
		if { $msc <= 0 } {
			return {}
		}
		thread::mutex lock $LOCK
		thread::cond wait $WAKEUP $LOCK $msc
		thread::mutex unlock $LOCK
		set msg [ tsv::lpop $TSVAR incoming ]
	}
	return $msg
}
##
# loop
# цикл опроса очереди
#
# разбирает входящую очередь
##
method loop { { timeout 0 } } {
	if { $timeout!=0 && $timeout < $startTime } {
		set timeout [ expr { $timeout + [ clock milliseconds ] } ]
	}
	set ret 0
	# timestamp type sender id args
	set msglist {}
	tsv::lock $TSVAR {
		set msglist [ tsv::get $TSVAR incoming ]
		tsv::set $TSVAR incoming {}
	}
	while { $ret == 0 } {
		# забираем все сообщения в очереди
		# разбираем все 
		foreach msg $msglist {
			if { [ llength $msg ] < 4 } {
				continue
			}
			lassign $msg msgtimeout type sender msgid
			if { $type == 1 } {
				# запрос к нам, пока не реализовано
				continue
			}
			if { ! [ info exists PROMISE($msgid) ] } {
				# нет у нас такого запроса
				continue
			}
			incr ret
			if { $type == 2 } {
				# успешное завершение
				set content [ lindex $msg 4 ]
				#if { [ llength $content ] == 1 } { set content [ lindex $content 0 ] }
				dict set PROMISE($msgid) result $content
			} else {
				# type 3 и прочие, как ошибки
				dict set PROMISE($msgid) error [ lindex $msg 4 ]
			}
		}
		if { $ret > 0 } {
			break
		}		
		if { $timeout == 0 } {
			# однократный проход 
			break
		}
		set mcs [ expr { $timeout - [ clock milliseconds ] } ]
		if { $mcs < 0 } {
			# вышел таймер
			break
		}
		if { $mcs > 50 } {
			# при большом таймере - вызвать update
			update
			update idletasks
			
			set mcs [ expr { $timeout - [ clock milliseconds ] } ]
			if { $mcs < 0 } {
				# вышел таймер (update долго исполнялся)
				# но сообщения всё-таки надо выгребать
				set mcs 2
			}
		}
		
		if { $mcs > 50 } {
			set mcs 50
		}
		
		try {
			thread::mutex lock $LOCK
			set msglist [ tsv::get $TSVAR incoming ]
			if { $msglist == {} } {
				thread::cond wait $WAKEUP $LOCK $mcs
				set msglist [ tsv::get $TSVAR incoming ]
			}
			thread::mutex unlock $LOCK
		} on error code {
			break
		}
		if { $msglist == {} } {
			break
		}
	}
	return $ret
}
method error { promise } {
	if { ! [ info exists PROMISE($promise) ] } {
		return {}
	}
	if { ! [ dict exists $PROMISE($promise) error ] } {
		return {}
	}
	dict get $PROMISE($promise) error 
}
method result { promise } {
	if { ! [ info exists PROMISE($promise) ] } {
		return {}
	}
	if { ! [ dict exists $PROMISE($promise) result ] } {
		return {}
	}
	dict get $PROMISE($promise) result 
}

method exists { promise } {
	if { ! [ info exists PROMISE($promise) ] } {
		return 0
	}
	return 1
}

method status { promise } {
	if { ! [ info exists PROMISE($promise) ] } {
		return -1
	}
	if { [ dict exists PROMISE($promise) error ] } {
		return 2
	}
	if { [ dict exists PROMISE($promise) result ] } {
		return 1
	}
	return 0
}

method timeout { promise } {
	if { ! [ info exists PROMISE($promise) ] } {
		return 0
	}
	dict get $PROMISE($promise) timeout
}
method print args {
	tsv::lpush $TSVAR incoming [ list 1 [ join $args " " ] ]
}
method alert args {
	tsv::lpush $TSVAR incoming [ list 2 [ join $args " " ] ]
}
method fault args {
	tsv::lpush $TSVAR incoming [ list 3 [ join $args " " ] ]
}
export id ready 
export complete await async asyncmany call invoke
export all any forget
export error result exists status timeout
};
