package provide MTWorker 0.1

package require Thread
package require uuid
package require lambda 
package require rl_json

oo::class create MTWorker {
variable TSVAR		; # имя переменной в tsv, по умолчанию "MTWorkers"
variable GUID		; # уникальный GUID = [ uuid::uuid generate ]

variable QUEUE		; # имя общей очереди, "incoming"
variable PRIVQ		; # имя собственной очереди "incoming/$GUID"
variable WAKEUP		; # thread::cond для оповещения о наличии входящих
variable LOCK 		; # блокирующий thread::mutex 

variable POOL 			; # пул для исполнения tcl,py

variable COMMANDS   ; # регестрированные команды

variable startTimeMsc;	# время собственного старта в милисек
variable startTime

constructor {} {
	set startTimeMsc [ clock milliseconds ]
	set startTime    [ expr { $startTimeMsc / 1000 } ]
	set TSVAR "MTWorkers" ; # имя констнтное - чтобы MTClient точно его находили
	set GUID 	[ uuid::uuid generate ]
	set QUEUE "incoming"
	set PRIVQ "incoming/$GUID"
	set POOL ""	
	array set COMMANDS {}
	# при старте в tsv будут созданны:
	# MTWorkers {
	#    incoming - очередь для входящих запросов
	#    wakeup   - tsv::для оповещений
	#    lock     - блокирующий mutex
	#    total    - кол-во созданных Workers
	#    ready    - кол-во ожидающих Workers
	#    pool     - tpool для исполнения tcl и py
	# }
	# для этого блокируем TSVAR и задаём нужную структуру
	tsv::lock $TSVAR {
		# создать собстенную очередь
		tsv::set $TSVAR $PRIVQ {}
		
		# общее кол-во Workers
		if { ! [ tsv::exists $TSVAR total ] || [ tsv::get $TSVAR total ] == 0 } {
			tsv::set $TSVAR total 1
			tsv::set $TSVAR ready 1
			# очередь входящих
			tsv::set $TSVAR incoming {}
			tsv::set $TSVAR wakeup [ thread::cond create ]
			tsv::set $TSVAR lock [ thread::mutex create ]
			if { ! [ tsv::exists $TSVAR counter ] || [ tsv::get $TSVAR counter ] < 10 } {
				tsv::set $TSVAR counter 10
			}
			# создание пула
			if { ! [ tsv::exists $TSVAR pool ] } {
				my init_pool
			}
		} else {
			tsv::incr $TSVAR total
			tsv::incr $TSVAR ready
		}
		# (закончили создавать pool)
		
		# и конечно сохранить у себя 
		set WAKEUP [ tsv::get $TSVAR wakeup ]
		set LOCK [ tsv::get $TSVAR lock ]
		set POOL [ tsv::get $TSVAR pool ]
		if { $POOL == "" } {
			alert "Unable to create Thread Pool"
		}		
	}
}
##
# init_pool
# инициализация пула tpool
# вызывается в конструкторе и только для самого первого воркера
# $TSVAR при этом залочен
# процедуры инициализации/деинита пула initcmd/exitcmd не должны
# иметь обращений к MTWorker
##
method init_pool {} {
	set initlist {}
	set exitlist {}
	# копирование массива ::MetaTrader
	if { [ array exists ::MetaTrader ] } {
		foreach { name value } [ array get ::MetaTrader ] {
			lappend initlist [ list array set ::MetaTrader [ list $name $value ] ]
		}
	}
	# пакеты "по умолчанию"
	lappend initlist { package require Thread }
	lappend initlist { package require lambda }
	lappend initlist { package require uuid }
	lappend initlist { package require rl_json }
	lappend initlist { catch { package require tclpy } }
	lappend initlist { package require http }
	lappend initlist { package require tls }
	lappend initlist { ::http::register https 443 [ list ::tls::socket -autoservername 1 ] }

	# копирование процедур
	foreach procname {
		work_print
		work_alert
		work_ok
		work_error
		work_tcl
		work_py
		work_exec
		work_bgexec
		work_start
		work_geturl
		work_mpi
	} {
		catch {
			set body [ info body $procname ]
			if { $body != {} } {
				lappend initlist [ list proc $procname [ info args $procname ] $body ]
			}
		}
	}
	# файлы инициализации
	set initfiles {}
	set exitfiles {}
	if { [ info exists ::MetaTrader(DataDir) ] } {
		lappend initfiles [ file join ::MetaTrader(DataDir) tcl lib mtworkers initcmd.tcl ]
		lappend exitfiles [ file join ::MetaTrader(DataDir) tcl lib mtworkers exitcmd.tcl ]
	} else {
		lappend initfiles [ file join [ pwd ] tcl lib mtworkers initcmd.tcl ]
		lappend exitfiles [ file join [ pwd ] tcl lib mtworkers exitcmd.tcl ]
	}
	if { [ info exists ::MetaTrader(ProgramDir) ] } {
		lappend initfiles [ file join ::MetaTrader(ProgramDir) initcmd.tcl ]
		lappend exitfiles [ file join ::MetaTrader(ProgramDir) exitcmd.tcl ]
	} else {
		lappend initfiles [ file join [ pwd ] initcmd.tcl ]
		lappend exitfiles [ file join [ pwd ] exitcmd.tcl ]
	}
	foreach initfile $initfiles {
		catch {
			set f [ open $init "r" ]
			fconfigure $f -encoding utf-8
			set content [ read $f ]
			close $f
			lappend initlist [ list catch [ list eval $content ] ]
		}
	}
	set exitfiles [ lreverse $exitfiles ]
	foreach exitfile $exitfiles {
		catch {
			set f [ open $init "r" ]
			fconfigure $f -encoding utf-8
			set content [ read $f ]
			close $f
			lappend exitlist [ list catch [ list eval $content ] ]
		}
	}	
	set initcmd [ join $initlist "\n" ]
#	set f [ open "zzz.tcl" w ] 
#	puts $f $initcmd
#	close $f
	set exitcmd [ join $exitlist "\n" ]
	set POOL [ tsv::set $TSVAR pool [ tpool::create -minworkers 1 -maxworkers 5 -initcmd $initcmd -exitcmd $exitcmd ] ]
}
destructor {
	# надо ещё отрабатывать общий деструктор
	# для этого блокируем MTWorkers
	tsv::lock $TSVAR {
		# если после декремента total, получается что нет никого
		# то выполняем общий деструктор
		if { [ tsv::incr $TSVAR total -1 ] <= 0 } {
			catch {
				set LOCK [ tsv::get $TSVAR lock ]
				thread::mutex destroy $LOCK
			}
			catch {
				set POOL [ tsv::get $TSVAR pool ]
				tpool::release $POOL
			}
			catch {
				set WAKEUP [ tsv::get $TSVAR wakeup ]
				thread::cond destroy $WAKEUP
			}
			tsv::unset $TSVAR lock
			tsv::unset $TSVAR pool
			tsv::unset $TSVAR wakeup
			tsv::unset $TSVAR incoming
			tsv::set $TSVAR total 0
			tsv::set $TSVAR ready 0
			
		} else {
			# иначе просто уменьшаем счётчик готовых
			tsv::unset $TSVAR $PRIVQ
			tsv::incr $TSVAR total -1
			tsv::incr $TSVAR ready -1
		}
	}
}

##
# message_correct msg
# проверить сообщение на корректность
# сообщение в виде : timeout type sender msgid ?content?
## 
method message_correct { msg } {
	set timeout [ lindex $msg 0 ]
	if { $timeout > 0 && $timeout < $startTime } {
		# "сквозное сообщение"
		return true
	}
	set len [ llength $msg ]
	if { $len != 5 } {
		# ровно 5 полей
		#puts stderr "message len incorrect"
		return false
	}
	# разбираем на поля
	lassign $msg timeout type sender msgid content
	# поля параноидально проверяем :-) 
	if { $timeout == {} || ! [ string is wideint $timeout ] || $timeout < 0 } {
		# таймаут - большое целое, положительное число
		#puts stderr "timeout incorrect"
		return false
	}
	if { $type == {} || ! [ string is int $type ] || $type <= 0 } {
		# тип - целое положительное число
		#puts stderr "type incorrect"
		return false
	}
	if { $sender == {} || ! [ tsv::exists $sender incoming ] } {
		# отправитель - должен существовать и иметь очередь для ответов
		#puts stderr "sender incorrect"
		return false
	}
	if { $msgid == {} || [ string is int $msgid ] && $msgid < 0 } {
		# если msgid число, то не меньшее 0 (0-внутренние сообщения)
		# (а вот строки могут быть произвольны)
		#puts stderr "msgid incorrect"
		return false
	}
	if { $type == 1 } {
		set command [ lindex $content 0 ]
		if { $command == {} || [ string trim $command ] != $command || [ llength $command ] != 1 } {
			# в запросах - если уж задана команда, то не пустая и из одного слова
			#puts stderr "command $command for type 1 incorrect"
			#exit
			return false
		}
	}
	return true
}
##
# getmessage ?timeout? 
# получить следующее сообщение из очереди входящих
# 	timeout - абс. время в милисек до которого можно ждать
# ----
# процедуру дёргает MetaTrader для получения сообщений
# возвращаем сообщение или {} если нет сообщений (или было отброшено/сразу_обработано)
##
method getmessage { {timeout 0} } {
	if { $timeout > 0 && $timeout < $startTime } {
		set timeout [ expr { [ clock milliseconds ] + $timeout } ]
	}
	# просто lpop
	# сначала смотрим "личную" очередь
	set msg [ tsv::lpop $TSVAR $PRIVQ ]
	# а потом общую
	if { $msg == {} } {
		set msg [ tsv::lpop $TSVAR $QUEUE ]
	}
	if { $msg == {} && $timeout > 0 } {
		# если ничего не получено 
		# то будем ожидать получения нового из общей очереди
		# msc - сколько допустимо ждать в милисек
		set msc [ expr { $timeout - [ clock milliseconds ] } ]
		if { $msc > 0 } {
			thread::mutex lock $LOCK
			if { [ tsv::llength $TSVAR $QUEUE ] == 0 } {
				thread::cond wait $WAKEUP $LOCK $msc
			}
			thread::mutex unlock $LOCK
			set msg [ tsv::lpop $TSVAR $QUEUE ]
			if { $msg == {} && [ clock milliseconds ] < $timeout } {
				# если после сигнала ничего нет, значит надо снова глянуть в личное
				set msg [ tsv::lpop $TSVAR $PRIVQ ]
			}
		}
	}
	# сразу проверить сообщения
	if { $msg != {} } {
		# проверка таймаута сообщений
		set msgtimeout [ lindex $msg 0 ]
		if { $msgtimeout > 0 && $msgtimeout < $startTime } {	
			# сковзное 
			return $msg
		}		
		if { ! [ my message_correct $msg ] } {
			# некорретные сразу отбрасываются
			my print "incorrect message [ lrange $msg 0 3 ]"
			return {}
		}
		lassign $msg msgtimeout type sender msgid payload
		if { $msgtimeout>0 && $msgtimeout < [ clock milliseconds ] } {
			# таймаут обработки сообщения
			#set sender [ lindex $msg 2 ]
			#set msgid [ lindex $msg 3 ]
			# отправим ответ о таймауте
			#my reply_timeout $msgtimeout $sender $msgid ""
			# а само сообщение выбросим
			#puts stderr "message expired $msg"
			# puts "getmessage: message expired"
			return {}
		}
		# запросы 1 с командами tcl,py - обработать самостоятельно
		set command [ lindex $payload 0 ]
		if { [ info exists COMMANDS($command) ] } {
			# заменяем имя на номер
			set num [ dict get $COMMANDS($command) num ]
			set payload [ lreplace $payload 0 0 $num ]
			# и отдаём наверх
			return [ lreplace $msg 4 4 $payload ]
		}
		if { $type == 1 } {
			switch -exact -- $command {
				@echo {
					my reply_ok $timeout $sender $msgid [ lrange $payload 1 end ]
					set msg {}
				}
				@microseconds {
					my reply_ok $timeout $sender $msgid [ clock microseconds ]
					set msg {}
				}
				geturl - 
				tcl -
				py -
				mpi -
				bgexec -
				start -
				exec {
					tsv::incr $TSVAR ready -1
					my to_pool $msg
					tsv::incr $TSVAR ready 1
					set msg {}
				}
				work_mpi -
				work_geturl -
				work_py -
				work_exec -
				work_bgexec -
				work_start -
				work_tcl -
				source -
				after {
					my reply_error $timeout $sender $msgid "command denied"
					set msg {}
				}
				default {
					if { [ info commands $command ] == $command || [ info procs $command ] == $command } {
						# команда интерпретатора
						tsv::incr $TSVAR ready -1
						my to_local $msg
						tsv::incr $TSVAR ready 1
						set msg {}
					} else {
						# системная команда
						set execok [ auto_execok $command ]
						if { $execok != {} } {
							set payload [ linsert $payload 0 "exec" ]
							set msg [ lreplace $msg 4 4 $payload ]		
							tsv::incr $TSVAR ready -1
							my to_pool $msg
							tsv::incr $TSVAR ready 1
							set msg {}
						}
					}
				}
			}
		}
		return $msg
	}
	return {}
}
##
#	register num command ?preprocess? ?postproc?
#	регестрировать команду для дальнейшей передачи в MetaTrader
#   num - номер для подстановки вместо имени (под которым команда известна выше)
#   command - имя команды
##
method register { num command } {
	if { [ info exists COMMANDS($command) ] } {
		return 0
	}
	dict set COMMANDS($command) num $num
	dict set COMMANDS($command) command $command
	return 1
}
##
# reply_ok 
# сформировать и отправить сообщение об успешном завершении
# 	timeout - абс.время в милисек таймаута
# 	sender  - кому посылать
# 	msgid   - id сообщения
## 
method reply_ok { timeout sender msgid result } {
	set msg [ list $timeout 2 $TSVAR $msgid $result ]
	
	tsv::lpush $sender "incoming" $msg
	
	catch {
		# wakeup может не быть у клиента
		# или можем попасть когда он разрушен
		# поэтому в catch
		set client_wakeup [ tsv::get $sender wakeup ]
		thread::cond notify $client_wakeup
	}
}
##
# reply_error
# сформировать и отправить сообщение об ошибке
# параметры те-же что у reply_ok, но timeout будет проигнорен
# 	timeout - абс.время в милисек таймаута
# 	sender  - кому посылать
# 	msgid   - id сообщения
## 
method reply_error { timeout sender msgid error } {
	set msg [ list 0 3 $TSVAR $msgid $error ]
	tsv::lpush $sender "incoming" $msg
	
	catch {
		set client_wakeup [ tsv::get $sender wakeup ]
		thread::cond notify $client_wakeup
	}
}

##
# reply_timeout
# сформировать и отправить сообщение о таймауте
# параметры те-же что у reply_ok, 
# но timeout будет переставлен к content,а content проигнорирован
# 	timeout - абс.время в милисек таймаута
# 	sender  - кому посылать
# 	msgid   - id сообщения
## 
method reply_timeout { timeout sender senderid content } {
	set msg [ list 0 4 $TSVAR $senderid $timeout [ clock milliseconds ] ]
	tsv::lpush $sender "incoming" $msg
	catch {
		set client_wakeup [ tsv::get $sender wakeup ]
		thread::cond notify $client_wakeup
	}
}

##
# to_pool msg
# обработать команду в пуле
## 
method to_pool { msg } {
	lassign $msg timeout type sender msgid payload
	set cmdline $payload
	set command [ lindex $payload 0 ]
	set work {}
	switch -exact -- $command {
		tcl -
		pool {
			set work [ list work_tcl $TSVAR $timeout $type $sender $msgid [ lreplace $cmdline 0 0 ] ]
		}
		python -
		py {
			set work [ list work_py  $TSVAR $timeout $type $sender $msgid [ lreplace $cmdline 0 0 py ] ]
		}
		mpi {
			set work [ list work_mpi $TSVAR $timeout $type $sender $msgid [ lreplace $cmdline 0 0 ] ]
		}
		exec {
			set work [ list work_exec  $TSVAR $timeout $type $sender $msgid [ lreplace $cmdline 0 0 ] ]
		}
		bgexec {
			set work [ list work_bgexec  $TSVAR $timeout $type $sender $msgid [ lreplace $cmdline 0 0 ] ]
		}
		start {
			set work [ list work_start  $TSVAR $timeout $type $sender $msgid [ lreplace $cmdline 0 0 ] ]
		}
		geturl {
			set work [ list work_geturl  $TSVAR $timeout $type $sender $msgid [ lreplace $cmdline 0 0 ] ]
		}
		default {
			set work [ list work_tcl  $TSVAR $timeout $type $sender $msgid $cmdline ]
		}
	}
	if { $work == ""  } {
		my reply_error $timeout $sender $msgid "internal error no_work"
		return
	}
	# отправляем в пул
	if { [ catch {
		if { $POOL == "" } {
			my reply_error $timeout $sender $msgid $err
		} else {
			tpool::post -nowait -detached $POOL $work
		}
	} err ] } {
		my alert "pool fail $err"
		my reply_error $timeout $sender $msgid $err
	} else {
		my print "job accepted to pool: $work"
		#tpool::wait $POOL $job
		#set result [ tpool::get $POOL $job ]
	}
}

method to_local { msg } {
	lassign $msg timeout type sender msgid payload
	if { [ catch {
		set result [ uplevel #0 $payload ]
	} err ] } {
		my reply_error $timeout $sender $msgid $err
	} else {
		my reply_ok $timeout $sender $msgid $result
	}	
}
##
# process
#   устарешее 
##
method process { msg } {
	set timestamp [ lindex $msg 0 ]
	set type [ lindex $msg 1 ]
	set sender [ lindex $msg 2 ]
	set id [ lindex $msg 3 ]
	set command [ lindex $msg 4 ]
	set args [ lrange $msg 5 end ]
	if { $type == 1 } {
		switch $command {
			1 -
			"echo" {
				tsv::lpush $sender incoming [ list [ clock milliseconds ] 2 $TSVAR $id {*}$args ]
				return 1
			}
			3 - 
			"microseconds" {
				tsv::lpush $sender incoming [ list [ clock milliseconds ] 2 $TSVAR $id {*}$args [ clock microseconds ] ]
			}
		}
	}
	return 0
}
##
# loop
#   устаревшее
##
method loop { {timeout 0} } {
	set start [ clock milliseconds ]
	if { $timeout != 0 } {
		if { $timeout < [ clock seconds ] } {
			set timeout [ expr { $timeout + $start } ]
		}
	}
	set result {}
	while 1 {
		set result [ my getmessage $timeout ]
		if { $result != {} } {
			if { [ my process $result ] } {
				; #puts "processed"
			} else {
				return $result
			}
		}
		if { $timeout != 0 && [ expr { $timeout < [ clock milliseconds ] } ] } {
			return -code error "0 loop timeout"
		}
	}
}
# почти фокус - для передачи внутренних сообщений в MetaTrader
# используются малые значения таймаута
##
# вывести сообщение в журнал MetaTrader
##
method print args {
	# вывести на печать
	tsv::lpush $TSVAR incoming [ list 1 [ join $args " " ] ]
}
##
# высветить Alert 
##
method alert args {
	# вывести на печать
	tsv::lpush $TSVAR incoming [ list 2 [ join $args " " ] ]
}
## 
# сообщить о критическом сбое
# (например MetaTrader может удалить инстанс,интерпретатор и стартовать их заново)
##
method fault args {
	tsv::lpush $TSVAR incoming [ list 3 [ join $args " " ] ]
}
export getmessage loop register
};

##################################
# процедуры для tpool
##################################

proc simple_mpi { data cmd } {
	#
	# запускаем 
	#	mpiexec -n 5 tclsh $script
	#	(чуть переделанный их mpich/examples)
	# и передаём на стандартный ввод массив $data
	# кусками по 1024 значения
	# запускаем
	try {
		set cmdline "|mpiexec -n 2 tclsh ${cmd}"
		set io [ open $cmdline RDWR ]
	} on error err {
		return -code error "pipe open failed $err"
	}
	# передаём список $data по кускам (построчно)
	set dataTotal [ llength $data ]
	set dataOffset 0
	try {
		while { $dataOffset < $dataTotal } {
			set portion [ lrange $data $dataOffset [ expr { $dataOffset + 1023 } ] ]
			if { [ llength $portion ] == 0 } {
				break
			}
			puts $io [ join $portion " " ]
			flush $io
			set count [ llength $portion ]
			incr dataOffset $count
		}
	} on error err {
		set reply {}
		catch { set reply [ read $io ] }
		set err [ catch { close $io } ] 
		return -code error "data transfer error $err $cmdline"
	}
	# пустая строка - признак конца данных
	puts $io "\n\n"
	flush $io
	
	# читаем stdout процесса и накапливаем в result
	set result {}
	set error {}
	try {
		while { ! [ eof $io ] } {
			try {
				set str [ gets $io ]
				if { $str != {} } {
					#puts "mpi: $str"
					lappend result $str
				} else {
					break
				}
			} on error err {
				set error "error on read $err"
				break
			}
		}
	} on error err {
		set error "another err $err"
	}
	# завершаем процесс
	try {
		close $io
	} on error err { 
		return -code error $err
	}
	# отдаём результат
	if { $error != {} } {
		return -code error $error
	}
	return [ join $result "\n" ]
}
##
# work_ok 
# сформировать и отправить сообщение об успешном завершении
#   TSVAR   - собственный ID
# 	timeout - абс.время в милисек таймаута
# 	sender  - кому посылать
# 	msgid   - id сообщения
## 
proc work_ok { TSVAR sender msgid result } {
	set msg [ list 0 2 $TSVAR $msgid $result ]
	tsv::lpush $sender "incoming" $msg
	catch {
		# wakeup может не быть у клиента
		# или можем попасть когда он разрушен
		# поэтому в catch
		set client_wakeup [ tsv::get $sender wakeup ]
		thread::cond notify $client_wakeup
	}
}
##
# work_error 
# сформировать и отправить сообщение об ошибке
#   TSVAR   - собственный ID
# 	timeout - абс.время в милисек таймаута
# 	sender  - кому посылать
# 	msgid   - id сообщения
## 
proc work_error { TSVAR sender msgid result } {
	set msg [ list 0 3 $TSVAR $msgid $result ]
	tsv::lpush $sender "incoming" $msg
	catch {
		# wakeup может не быть у клиента
		# или можем попасть когда он разрушен
		# поэтому в catch
		set client_wakeup [ tsv::get $sender wakeup ]
		thread::cond notify $client_wakeup
	}
}
##
# work_print
# печать из пула
##
proc work_print { TSVAR sender msgid args } {
	set msg [ list 1 $args ]
	tsv::lpush $TSVAR "incoming" $msg
	catch {
		# wakeup может не быть у клиента
		# или можем попасть когда он разрушен
		# поэтому в catch
		set wakeup [ tsv::get $TSVAR wakeup ]
		thread::cond notify $wakeup
	}
}
##
# work_print
# alert из пула
##
proc work_alert { TSVAR sender msgid args } {
	set msg [ list 2 $args ]
	tsv::lpush $TSVAR "incoming" $msg
	catch {
		# wakeup может не быть у клиента
		# или можем попасть когда он разрушен
		# поэтому в catch
		set wakeup [ tsv::get $TSVAR wakeup ]
		thread::cond notify $wakeup
	}
}
##
# work_tcl
# исполнить задачу "tcl"
##
proc work_tcl { TSVAR timeout type sender msgid command } {
	try {
		set result [ uplevel #0 $command ]
		work_ok $TSVAR $sender $msgid $result
	} on error err {
		work_error $TSVAR $sender $msgid $err
	}
}
##
# work_py
# исполнить задачу "py"
# (в случае любых ошибков - рестарт треда)
##
proc work_py { TSVAR timeout type sender msgid pycommand } {
	try {
		set result [ uplevel #0 $pycommand ]
		work_ok $TSVAR $sender $msgid $result
	} on error err {
		work_error $TSVAR $sender $msgid $err
		thread::exit
	}
}
##
# work_exec 
# выполнить команду OC, stdout отдать как результат
# в случае ошибок - рестарт треда
##
proc work_exec { TSVAR timeout type sender msgid command } {
	try {
		work_print $TSVAR $sender $msgid "work_exec $command"
		set result [ exec -ignorestderr -- {*}$command ]
		work_print $TSVAR $sender $msgid "work_exec ok"
		work_ok $TSVAR $sender $msgid $result
	} on error err {
		work_print $TSVAR $sender $msgid "work_exec error $err"
		work_error $TSVAR $sender $msgid $err
		thread::exit
	}
}

##
# work_bgexec 
# выполнить команду OC, stdout отдать в workers в печать
# в случае ошибок - рестарт треда
##
proc work_bgexec { TSVAR timeout type sender msgid command } {
	try {
		lassign [ chan pipe ] rd wr
		fileevent $rd readable [ lambda { TSVAR sender msgid rd } {
			catch {
				if { ! [ eof $rd ] } {
					set str [ gets $rd ]
					work_print $TSVAR $sender $msgid $str
					if { [ eof $rd ] } {
						catch { close $rd }
					}
				}
			}
		} $TSVAR $sender $msgid $rd ]
		set result [ exec -ignorestderr -- {*}$command 2>@1 1>@ $wr ]
		work_ok $TSVAR $sender $msgid 0
	} on error err {
		work_error $TSVAR $msgid $err
		thread::exit
	}
}
##
#
##
proc work_start { TSVAR timeout type sender msgid command } {
	try {
		set pids [ exec -ignorestderr -- cmd /c start "$sender $msgid" {*}$command > NUL 2>@1 ]
		work_ok $TSVAR $sender $msgid $pids
	} on error err {
		work_error $TSVAR $msgid $err "work_start error $err"
		thread::exit
	}
}
##
# work_mpi
# выполнить команду mpiexec command args..
# и отправить туда данные на стандартный ввод, нарезая построчко кусками по strides
# в случае ошибок - рестарт треда
##
proc work_mpi { TSVAR timeout type sender msgid payload } {
	try {
	
	set data {}
	set size 0
	set binFormat {}
	set strides 1
	set sep1 " "
	set sep2 " "
	# забрать аргументы
	while { [ llength $payload ] > 0 } {
		if { [ lindex $payload 0 ] == "-tsv" } {
			set var [ lindex $payload 1 0 ]
			set key [ lindex $payload 1 1 ]
			if { ! [ tsv::exists $var $key ] } {
				work_alert $TSVAR $sender $msgid "data no exists in tsv { $var $key }"
				work_error $TSVAR $sender $msgid "data no exists in tsv { $var $key }"
				return
			}
			if { [ catch {
				set data [ tsv::get $var $key ]
			} err ] } {
				work_alert $TSVAR $sender $msgid "tsv get $var $key : $err"
				work_error $TSVAR $sender $msgid "tsv get $var $key : $err"
				return
			}
			set payload [ lrange $payload 2 end ]
			continue
		}
		if { [ lindex $payload 0 ] == "-data" } {
			set data [ lindex $payload 1 ]
			set payload [ lrange $payload 2 end ]
			continue
		}
		if { [ lindex $payload 0 ] == "-format" || [ lindex $payload 0 ] == "-bin" } {
			set binFormat [ lindex $payload 1 ]
			set payload [ lrange $payload 2 end ]
			continue
		}
		if { [ lindex $payload 0 ] == "-strides" } {
			set strides [ lindex $payload 1 ]
			set payload [ lrange $payload 2 end ]
			continue
		}
		break
	}
	# проверить аргументы
	if { $strides == {} || ! [ string is int $strides ] || $strides <= 0 } {
		work_error $TSVAR $sender $msgid "-strides {int>0}"
		return
	}
	if { $binFormat == {} } {
		# если данные текстовые
		set size [ llength $data ]
		if { $size > 0 } {
			if { [ expr { $size % $strides } ] != 0 } {
				work_error $TSVAR $sender $msgid "-data \[%strides\]"
				return
			}
		}
	} else {
		if { [ llength $binFormat ] != 1 } {
			work_error $TSVAR $sender $msgid "-bin no_space_record"
			return
		}
		set size [ string length $data ]
	}
	# формируем  ком.строку
	set cmdname [ string trim [ lindex $payload 0 ] ]
	set cmdline {}
	if { $cmdname == {} } {
		work_error $TSVAR $sender $msgid "empty command"
		return
	}
	set ext [ string tolower [ file extension $cmdname ] ]
	if { $ext == ".tcl" || $ext == ".py" } {
		if { ! [ file readable $cmdname ] || ! [ file isfile $cmdname ] } {
			work_error $TSVAR $sender $msgid "$cmdname not readable"
			return
		}
		if { $ext == ".tcl" } {
			set interp "tclsh"
		} else {
			set interp "python"
		}
		set cmdline [ list $interp {*}$payload ]
		set cmdname $interp
	}
	flush stdout
	# проверить возможность запуска
	set execok [ auto_execok $cmdname ]
	if { $execok == {} } {
		work_error $TSVAR $sender $msgid "$cmdname not executable"
		return
	}
	#set cmdname [ lreplace $cmdname 0 0 $execok ]
	set cmdline [ linsert $cmdline 0 [ auto_execok "mpiexec" ] -n 5 ]
	# будем читать выхлоп и складывать в result
	global workResult
	global workCode
	
	set workResult {}
	set workCode "-"
	
	set reader [ lambda { TSVAR sender msgid rd } {
		try {
			global workResult
			global workCode
			if { [ eof $rd ] || [ catch { set str [ gets $rd ] } ] } {
				# конец файла или ошибка чтения
				set workCode [ catch { close $rd } ]
				return
			}
			if { $str != {} } {
				lappend workResult $str
				work_print $TSVAR $sender $msgid "work_mpi:$str"
			} else {
				set workCode 0
				#catch { close $rd }
				#upvar wr wr
				#catch { close $wr }
				#return
			}
			if { [ eof $rd ] } {
				set workCode [ catch { close $rd } ]
			}
		} on error err {
			work_alert $TSVAR $sender $msgid "lambda error $err"
		}
	} $TSVAR $sender $msgid ]
	# проверим наличие smpd
	set have_smpd 0
	set smpd ""
	try {
		set stat [ exec -- netstat -a -n -p tcp ]
		foreach line [ split $stat \n ] {
			if { [ lindex $line 0 ] == "TCP" && [ lindex $line end ] == "LISTENING" } {
				set port [ lindex [ split [ lindex $line 1 ] ":" ] 1 ]
				if { $port == 8676 } {
					# smpd may be run
					set have_smpd 1
					break
				}
			
			}
		}
		if { ! $have_smpd } {
			work_alert $TSVAR $sender $msgid "smpd not started as service, run locally in debug mode(much slow)"
			set smpd [ open "|smpd -p 8676 -d > NUL 2>@1" WRONLY ]
		} else {
			work_print $TSVAR $sender $msgid "smpd already run"
		}
	} on error err {
		work_alert $TSVAR $sender $msgid "internal error $err"
	}
	
	# готовим пайп для чтения stdout процесса
	lassign [ chan pipe ] rd wr
	fconfigure $rd -buffering line -blocking 1
	fconfigure $wr -buffering line -blocking 1
	fileevent $rd readable [ list {*}$reader $rd ]
	# готовим для записи в stdin процесса
	lassign [ chan pipe ] rd1 wr1
	fconfigure $rd1 -buffering line -blocking 1
	fconfigure $wr1 -buffering line -blocking 1
	# запускаем
	try {
		set pids [ exec -ignorestderr -- {*}$cmdline <@ $rd1 >@ $wr & ]
	} on error err {
		work_alert $TSVAR $sender $msgid "$cmdline : $err"
		catch { close $rd }
		catch { close $wr }
		catch { close $rd1 }
		catch { close $wr1 }
		return 
	}
	# если $data не пустое, то складываем туда (wr1)
	# данные построчно кусками по strides
	# и завершаем пустой строкой
	if { $size > 0 } {
		#work_print $TSVAR $sender $msgid "transfer data"
		if { $binFormat == {} } {
			# текстовые данные
			set offset 0
			while { $offset < $size } {
				set portion [ join [ lrange $data $offset [ expr { $offset + $strides - 1 } ] ] $sep1 ]
				if { $portion!= "" } {
					puts $wr1 [ join $portion $sep2 ]
					flush $wr1
				}
				incr offset [ llength $portion ]
			}
			# пустая строка
			puts $wr1 "\n"
			flush $wr1
		} else {
			# бинарные
			# определяем размер одной записи
			set value {}
			binary scan $data $binFormat value
			set reclen [ string length [ binary format "${binFormat}" $value ] ]
			#work_print $TSVAR $sender $msgid "record size $reclen bytes"

			set end 0
			set stride {}
			set offset 0
			set count 0
			while { !$end && $offset < $size } {
				set stride {}
				set record [ string range $data $offset [ expr { $offset + $reclen - 1 } ] ]
				set value ""
				if { [ binary scan $record $binFormat value ] != 1 || $value =="" } {
					set end 1
				}
				incr offset $reclen
				if { $value!= "" } {
					lappend stride [ join $value $sep1 ]
				}
				if { $end || [ llength $stride ] == $strides } {
					puts $wr1 [ join $stride $sep2 ]
					set stride {}
				}
			}
			puts $wr1 "\n"
			flush $wr1
		}
		#work_print $TSVAR $sender $msgid "data transferred"
	}
	# ожидаем кода завершения
	while { ! [ eof $wr ] && ! [ eof $rd1 ] && $workCode=="-"} {
		after 20
		update
		update idletasks
	}
	work_print $TSVAR $sender $msgid "calc done"
	#vwait code
	flush stdout
	catch { close $rd }
	catch { close $wr }
	catch { close $rd1 }
	catch { close $wr1 }
	if { $smpd !="" } {
		set smpd_pid [ pid $smpd ]
		catch { exec cmd /c taskkill /PID $smpd_pid /T /F }
		catch { close $smpd } err
	}
	work_print $TSVAR $sender $msgid "work_mpi end"
	if { $workCode == 0 } {
		# успешное завершение
		work_print $TSVAR $sender $msgid "work_mpi ok"
		work_ok $TSVAR $sender $msgid $workResult
	} else {
		work_print $TSVAR $sender $msgid "work_mpi error"
		work_error $TSVAR $sender $msgid "code=$workCode"
	}

	} on error err {
		work_alert $TSVAR $sender $msgid "internal work_mpi error $err"
	}
	
}

proc work_geturl { TSVAR timeout type sender msgid payload } {
	try {

	# разбор аргументов
	set headers {}
	set method [ list -method GET ]
	set validate 1
	set strict 1
	set postproc {}
	set detail 0
	set repeats 5
	set interval 1000
	set url [ lindex $payload 0 ]
	set payload [ lrange $payload 1 end ]
	while { $payload != {} } {
		if { [ lindex $payload 0 ] == "-headers" } {
			set headers [ list -headers [ lindex $payload 1 ] ]
			set payload [ lrange $payload 2 end ]
			continue
		}
		if { [ lindex $payload 0 ] == "-method" } {
			set method [ list -method [ lindex $payload 1 ] ]
			set payload [ lrange $payload 2 end ]
			continue
		}
		if { [ lindex $payload 0 ] == "-strict" } {
			set strict [ list -strict [ lindex $payload 1 ] ]
			set payload [ lrange $payload 2 end ]
			continue
		}
		if { [ lindex $payload 0 ] == "-postproc" } {
			set postproc [ lindex $payload 1 ]
			set payload [ lrange $payload 2 end ]
			continue
		}
		if { [ lindex $payload 0 ] == "-detail" } {
			set detail [ lindex $payload 1 ]
			set payload [ lrange $payload 2 end ]
			continue
		}
		
		break
	}
	if { $url == "" } {
		work_error $TSVAR $sender $msgid "work_geturl url empty"
		return
	}
	#work_print $TSVAR $sender $msgid "work_geturl http start $url "
	# несколько попыток на получить ответ http
	for { set attempt 1 } { $attempt <= $repeats } { incr attempt } {
		set htt ""
		if { $timeout !=0 } {
			set httptimeout [ expr { $timeout - [ clock milliseconds ] } ]
			if { $httptimeout < 0 } {
				work_error $TSVAR $sender $msgid "work_geturl job timeout"
				return
			}
			set htt "-timeout $httptimeout"
		}
		try {
			set token [ http::geturl $url {*}$method {*}$headers {*}$htt ]
		} on error err {
			#work_print $TSVAR $sender $msgid "work_geturl attempt $attempt failed :$err"
			if { $attempt == $repeats } {
				work_error $TSVAR $sender $msgid "work_geturl maximum attempts reached"
				catch { http::cleanup $token }
				return
			}
			if { $timeout != 0 && $timeout < [ clock milliseconds ] } {
				work_error $TSVAR $sender $msgid "work_geturl timeout"
				catch { http::cleanup $token }
				return
			}
			incr attempt
			after $interval
			continue
		}
		break
	}
	#work_print $TSVAR $sender $msgid "work_geturl http complete $url"

	# разобрать результат
	upvar $token state
	if { [ info exists state(error) ] } {
		work_error $TSVAR $sender $msgid "work_geturl http error: $state(error)"
		catch { http::cleanup $token }
		return
	}
	set ncode [ http::ncode $token ]
	if { $ncode != 200 } {
		work_error $TSVAR $sender $msgid [ list $ncode [ array get state ] ]
		catch { http::cleanup $token }
		return
	}
	if { $detail } {
		set content [ array get state ]
	} elseif { [ info exists state(body) ] } {
		set content $state(body)
	} else {
		set content {}
	}
	catch { http::cleanup $token }
	#work_print $TSVAR $sender $msgid "work_geturl http body [ lindex [ split $content \n ] 0 ] ..."
	
	# если надо - исполнить постпроцесс
	if { $postproc == {} } {
		set result $content
	} else {
		try {
			set result [ {*}$postproc $content ]
		} on error err {
			work_error $TSVAR $sender $msgid "work_geturl posprocess error: $err"
			return
		}
	}

	#work_print $TSVAR $sender $msgid "work_geturl job done $url"
	work_ok $TSVAR $sender $msgid $result
	
	} on error err {
		#work_alert $TSVAR $sender $msgid "work_geturl internal error: $err"
		work_error $TSVAR $sender $msgid "work_geturl internal error: $err"
	}
}
